//  Sample Delphi 3 program to rotate a 24-bit color bitmap any angle.
//
//  A 640 x 480 pixel x 24-bit color bitmap can be rotated any angle
//  in about a second on a 166 MHz Pentium.
//
//  Copyright (C) 1997-1999 Earl F. Glynn, Overland Park, KS  USA.
//  All Rights Reserved.

unit gRotate;

interface

uses
  Windows, Messages, SysUtils, Classes,
  Graphics, Controls, Forms, Dialogs, Math,
  ExtDlgs, StdCtrls, ExtCtrls, Spin, Buttons;

type
  TRotateForm = class(TForm)
    Panel1: TPanel;
    ScrollBox1: TScrollBox;
    ImageRotated: TImage;
    Panel2: TPanel;
    SpinEditI: TSpinEdit;
    LabelI: TLabel;
    LabelCenter: TLabel;
    LabelJ: TLabel;
    SpinEditJ: TSpinEdit;
    SpeedButton1: TSpeedButton;
    SpinEditThetaDegrees: TSpinEdit;
    Label1: TLabel;
    SpinEditThetaDegreesHundredths: TSpinEdit;
    ButtonSaveRotated: TButton;
    LabelDegreesClockwise: TLabel;
    HelpButton: TSpeedButton;
    ExitButton: TSpeedButton;
    procedure SpinEditRotate(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure ButtonSaveRotatedClick(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure FormShow(Sender: TObject);
    procedure SpeedButton1Click(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure HelpButtonClick(Sender: TObject);
    procedure ExitButtonClick(Sender: TObject);

  private
    KeepIt,
      Busy: BOOLEAN; // Flag to avoid multiple rotates when reading
                                // image and setting center of rotation
{    BitmapOriginal:  TBitmap;}
  public
    procedure RotateBitmap;
  end;

var
  RotateForm: TRotateForm;

implementation
{$R *.DFM}
uses {UAO,}  gGlobal, gREIV;

procedure TRotateForm.FormCreate(Sender: TObject);
begin
  left := RotateFormX;
  top := RotateFormY;
  Busy := TRUE;
  KeepIt := False;
end;

procedure TRotateForm.FormCloseQuery(Sender: TObject;
  var CanClose: Boolean);
begin
  if Busy then CanClose := false;
end;

procedure TRotateForm.FormClose(Sender: TObject; var Action:
  TCloseAction);
begin
  RotateFormX := RotateForm.left;
  RotateFormY := RotateForm.top;
  DoSaver;
end;

procedure TRotateForm.HelpButtonClick(Sender: TObject);
begin
  Application.HelpContext(5555);
end;

procedure TRotateForm.ExitButtonClick(Sender: TObject);
begin
  Close;
end;

// Load 24-bit color image from Main Raster

procedure TRotateForm.FormShow(Sender: TObject);
begin
  if REIVForm.Image1.Picture.Bitmap <> nil then
  begin
    Busy := TRUE; // avoid extra rotates caused by these assignments
    SpinEditI.Value := REIVForm.Image1.Picture.Bitmap.Width div 2;
    SpinEditJ.Value := REIVForm.Image1.Picture.Bitmap.Height div 2;
    Busy := FALSE;
    // Rotate and display the image
    RotateBitmap
  end else Busy := TRUE;
end;

// Rotate image by angle[degrees] specified in spinbox

procedure TRotateForm.SpinEditRotate(Sender: TObject);
begin
  if not Busy then RotateBitmap
end;


// Use center of pixel (i + 0.5, j + 0.5) in the "reverse" lookup.

function RotateBitmapMethod2({CONST BitmapOriginal:  TBitmap;}
  const iRotationAxis, jRotationAxis: INTEGER;
  const AngleOfRotation: DOUBLE; // radians
  var RotateMilliseconds: DWORD): TBitmap;
type
  pRGBTripleArray = ^TRGBTripleArray;
  TRGBTripleArray = array[0..65535] of TRGBTriple;

var
  cosTheta: Single; // Use Single for speed
  i: INTEGER;
  iOriginal: INTEGER;
  xPrime: SINGLE;
  j: INTEGER;
  jOriginal: INTEGER;
  yPrime: SINGLE;
  RowOriginal: pRGBTRipleArray;
  RowRotated: pRGBTRipleArray;
  sinTheta: Single;
  StartTime: DWORD;
  Temp1, Temp2: EXTENDED;
begin
  // The size of BitmapRotated is the same as BitmapOriginal.PixelFormat
  // must also match since 24-bit GBR triplets are assumed in ScanLine.
  RESULT := TBitmap.Create;
  RESULT.Width := REIVForm.Image1.Picture.Bitmap.Width;
  RESULT.Height := REIVForm.Image1.Picture.Bitmap.Height;
  RESULT.PixelFormat := pf24bit; // Force This

  // "Start" the clock
  StartTime := GetTickCount;

  // Get SIN and COS in single call from math library
  sincos(AngleOfRotation, Temp1, Temp2);
  sinTheta := Temp1; // kludge around type conversion
  cosTheta := Temp2;

// If no math library, then use this:
// sinTheta := SIN(AngleOfRotation);
// cosTheta := COS(AngleOfRotation);

  // Step through each row of rotated image.
  for j := RESULT.Height - 1 downto 0 do
  begin
    RowRotated := RESULT.Scanline[j];
    yPrime := (j - jRotationAxis) + 0.5; // use center of Pixel

    for i := RESULT.Width - 1 downto 0 do
    begin
      xPrime := (i - iRotationAxis) + 0.5; // use center of pixel

      iOriginal := iRotationAxis + ROUND(xPrime * CosTheta - yPrime *
        sinTheta - 0.5);
      jOriginal := jRotationAxis + ROUND(xPrime * sinTheta + yPrime *
        cosTheta - 0.5);

      // Make sure (iOriginal, jOriginal) is in BitmapOriginal.  If not,
      // assign blue color to corner points.
      if (iOriginal >= 0) and (iOriginal <=
        REIVForm.Image1.Picture.Bitmap.Width - 1) and
        (jOriginal >= 0) and (jOriginal <=
        REIVForm.Image1.Picture.Bitmap.Height - 1)
        then begin
        // Assign pixel from rotated space to current pixel in BitmapRotated
        RowOriginal :=
          REIVForm.Image1.Picture.Bitmap.Scanline[jOriginal];
        RowRotated[i] := RowOriginal[iOriginal]
      end
      else begin
        RowRotated[i].rgbtBlue := 255; // assign "corner" color
        RowRotated[i].rgbtGreen := 0;
        RowRotated[i].rgbtRed := 0
      end

    end
  end;
  RotateMilliseconds := GetTickCount - StartTime; // "stop" the clock
end {RotateBitmapMethod3};





procedure TRotateForm.RotateBitmap;
var
  BitmapRotated: TBitmap;
  iRotationAxis: INTEGER;
  jRotationAxis: INTEGER;
  Milliseconds: DWORD;
  Theta: DOUBLE;
begin
  try
    iRotationAxis := SpinEditI.Value
  except
    iRotationAxis := 0
  end;

  try
    jRotationAxis := SpinEditJ.Value
  except
    jRotationAxis := 0
  end;

  // Convert degrees to radians.  Use minus sign to force clockwise rotation.
  Theta := -(SpinEditThetaDegrees.Value +
    SpinEditThetaDegreesHundredths.Value / 100) *
    PI / 180;

  BitmapRotated :=
    RotateBitmapMethod2({RasterForm.Image1.Picture.Bitmap,}
    iRotationAxis, jRotationAxis,
    Theta, Milliseconds);

  try
    RotateForm.Caption := 'RotataRama Rotation Time = ' +
      IntToStr(Milliseconds) + ' ms';
    ImageRotated.Picture.Graphic := BitmapRotated;
    if KeepIt then
    begin
      REIVForm.Image1.Picture.Bitmap := BitmapRotated;
      KeepIt := False;
    end;
  finally
    BitmapRotated.Free
  end
end {RotateImage};


procedure TRotateForm.ButtonSaveRotatedClick(Sender: TObject);
begin
{Replace Raster image}
  KeepIt := True;
  RotateBitmap;
end;


procedure TRotateForm.SpeedButton1Click(Sender: TObject);
begin
  RotateBitmap;
end;





end.
